﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
FrequencyLoop v1 by MacGruber.
Variant of the Loop brick where the time between triggers is defined by a frequency rather than time. Also allows accelerating/deccelerating.

Version 1 - 2023-03-01
	Initial release.
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class FrequencyLoop : MVRScript
    {
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableFloat myTime;
		private JSONStorableBool myAddEndTick;
		private JSONStorableBool myAddEndWait;
		private JSONStorableFloat myFreqStart;
		private JSONStorableFloat myFreqEnd;
		private JSONStorableBool myFreqQuadratic;
		private EventTrigger myTickTrigger;
		private EventTrigger myFinishedTrigger;

		private int myState = STATE_DONE;
		private float myClock = -1.0f;
		private float myPrevTick = -1.0f;
		private float myNextTick = -1.0f;
		
		private const int STATE_DONE = 0;
		private const int STATE_RUNNING = 1;
		private const int STATE_ENDTICK = 2;
		private const int STATE_ENDWAIT = 3;

        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>FrequencyLoop</b></size>\nLogicBrick that triggers multiple times with at an accelerating or deccelerating frequency after " +
				"<i>Trigger</i> is called.\n\nCan also be used for constant frequency by setting start/end to the same value.\n\n" + 
				"<b>Add extra Tick</b> extends <i>Time</i> by rounding up to the next tick. This ensures the last tick actually reached <i>Frequency End</i>.\n\n" + 
				"<b>Wait after last Tick</b> adds some additional wait time before firing <i>OnFinishedTrigger</i>, this is based on <i>Frequency End</i>. Otherwise " +
				"<i>OnFinishedTrigger</i> is fired immediately after the last tick.\n\n" + 
				"<b>Use Quadratic Interpolation</b> enables quadratic instead of linear interpolation between <i>Frequency Start</i> and <i>Frequency End</i>.</color>",
				845.0f, true
			);
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);

			myTime = Utils.SetupSliderFloatWithRange(this, "Time (sec)", 10, 0, 100, false);
			myTime.constrained = false;
			myTime.setCallbackFunction += (float v) => { OnSettingsChanged(); };
			
			myAddEndTick = Utils.SetupToggle(this, "Add extra Tick", true, false);
			myAddEndTick.setCallbackFunction += (bool v) => { OnSettingsChanged(); };
			
			myAddEndWait = Utils.SetupToggle(this, "Wait after last Tick", true, false);
			myAddEndWait.setCallbackFunction += (bool v) => { OnSettingsChanged(); };
			
			myFreqStart = Utils.SetupSliderFloatWithRange(this, "Frequency Start (Hz)", 1, 0, 20, false);
			myFreqStart.constrained = false;
			myFreqStart.setCallbackFunction += (float v) => { OnSettingsChanged(); };
			
			myFreqEnd = Utils.SetupSliderFloatWithRange(this, "Frequency End (Hz)", 10, 0, 20, false);
			myFreqEnd.constrained = false;
			myFreqEnd.setCallbackFunction += (float v) => { OnSettingsChanged(); };
			
			myFreqQuadratic = Utils.SetupToggle(this, "Use Quadratic Interpolation", true, false);		
			myFreqQuadratic.setCallbackFunction += (bool v) => { OnSettingsChanged(); };

			myTickTrigger = new EventTrigger(this, "OnTickTrigger");
			Utils.SetupButton(this, "Actions OnTickTrigger", myTickTrigger.OpenPanel, false);
			
			myFinishedTrigger = new EventTrigger(this, "OnFinishedTrigger");
			Utils.SetupButton(this, "Actions OnFinishedTrigger", myFinishedTrigger.OpenPanel, false);

			Utils.SetupButton(this, "Test Trigger", Trigger, false);
			Utils.SetupAction(this, "Trigger", Trigger);
			
			Utils.SetupAction(this, "SetStartToEndFrequency", SetStartToEndFrequency);
			Utils.SetupAction(this, "SwapFrequency", SwapFrequency);
			Utils.SetupAction(this, "SetFinished", SetFinished);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myTickTrigger.SyncAtomNames();
			myFinishedTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myTickTrigger.Name] = myTickTrigger.GetJSON(base.subScenePrefix);
				jc[myFinishedTrigger.Name] = myFinishedTrigger.GetJSON(base.subScenePrefix);
				jc["Clock"].AsFloat = myClock;
				jc["PrevTick"].AsFloat = myPrevTick;
				jc["NextTick"].AsFloat = myNextTick;
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			Utils.AdjustSliderRange(myTime);
			myClock = -1.0f;
			myNextTick = -1.0f;

			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myTickTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myFinishedTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				
				myClock = jc["Clock"].AsFloat;
				myPrevTick = jc["PrevTick"].AsFloat;
				myNextTick = jc["NextTick"].AsFloat;
			}
		}
		
		private void OnDisable()
		{
			myState = STATE_DONE;
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myTickTrigger.Remove();
			myFinishedTrigger.Remove();
		}

		private void Update()
		{
			myTickTrigger.Update();
			myFinishedTrigger.Update();
			
			if (myState == STATE_DONE)
			{
				return;
			}
			
			myClock += Time.deltaTime;
			
			if (myState == STATE_RUNNING)
			{			
				if (myClock >= myTime.val)
				{			
					myState = STATE_ENDTICK;
				}
				else if (myClock >= myNextTick)
				{
					UpdateNextTick();
					myPrevTick = myClock;
					myTickTrigger.Trigger();
				}		
			}
			
			if (myState == STATE_ENDTICK)
			{
				if (!myAddEndTick.val)
				{
					myState = STATE_ENDWAIT;
					UpdateNextTick();
				}
				else if (myClock >= myNextTick)
				{
					myState = STATE_ENDWAIT;
					UpdateNextTick();
					myPrevTick = myClock;
					myTickTrigger.Trigger();
				}	
			}
			
			if (myState == STATE_ENDWAIT)
			{		
				if (!myAddEndWait.val || myClock >= myNextTick)
				{
					myState = STATE_DONE;
					myFinishedTrigger.Trigger();
				}		
			}
		}
		
		private void OnSettingsChanged()
		{
			if (myState != STATE_DONE)
			{
				myState = STATE_RUNNING;
				myNextTick = myPrevTick;
				UpdateNextTick();
				myNextTick = Mathf.Max(myNextTick, myClock);
			}
		}
		
		private void UpdateNextTick()
		{			
			float freqStart = Mathf.Max(myFreqStart.val, 0.01f);
			float freqEnd = Mathf.Max(myFreqEnd.val, 0.01f);
			float t = Mathf.Clamp01(myClock / myTime.val);
			if (myFreqQuadratic.val)
			{
				if (freqStart > freqEnd)
					t = 1.0f - Square(1.0f - t);
				else
					t = Square(t);
			}
			
			float frequency = Mathf.LerpUnclamped(freqStart, freqEnd, t);				
			myNextTick += 1.0f / frequency;
		}

		private void Trigger()
		{
			if (myAcceptTrigger.val)
			{
				myState = STATE_RUNNING;
				myClock = 0.0f;
				myPrevTick = 0.0f;
				myNextTick = 0.0f;
			}
		}
		
		
		private void SetStartToEndFrequency()
		{
			myFreqStart.val = myFreqEnd.val;
		}
		
		private void SwapFrequency()
		{
			float tmp = myFreqEnd.val;
			myFreqEnd.valNoCallback = myFreqStart.val;
			myFreqStart.val = tmp;
		}
		
		private void SetFinished()
		{
			myState = STATE_DONE;
			myFinishedTrigger.Trigger();
		}
		
		private static float Square(float x)
		{
			return x*x;
		}
    }
}
