/* /////////////////////////////////////////////////////////////////////////////////////////////////
LinearAnimation v1 by MacGruber.
Simple plugin that allows a linear animation between two points for the main control of an atom.

Version 1 - 2023-03-01
	Initial release.
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
	public class RandomAnimation : MVRScript
	{
		private JSONStorableBool myActive;
		private JSONStorableFloat myRadius;
		private JSONStorableFloat myRotationStrength;
		private JSONStorableFloat mySpeedMin;
		private JSONStorableFloat mySpeedMax;
		
		private Vector3 myAnchorPosition;
		private Quaternion myAnchorRotation;
		private Vector3 myVelocity;
		private Vector3 myTargetPosition;
		private Quaternion myTargetRotation = Quaternion.identity;
		private float myTargetTime;
		private float myClock;
		
		
		public override void Init()
		{
			ResetAnchor();
			
			myActive = Utils.SetupToggle(this, "Active", true, false);
			Utils.SetupButton(this, "Reset Anchor Position", ResetAnchor, false);
			
			myRadius = Utils.SetupSliderFloat(this, "Radius", 0.05f, 0.0f, 0.3f, false);
			myRotationStrength = Utils.SetupSliderFloat(this, "Rotation Strength", 0.1f, 0.0f, 0.5f, false);
			mySpeedMin = Utils.SetupSliderFloat(this, "Speed Min (cm/s)", 0.5f, 0.01f, 5.0f, false);
			mySpeedMax = Utils.SetupSliderFloat(this, "Speed Max (cm/s)", 1.0f, 0.01f, 5.0f, false);
			
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>RandomAnimation v0.1</b></size>\nSimple plugin provides a random motion within a sphere for the main control of an atom.</color>",
				200.0f, true
			);
		}
		
		private void ResetAnchor()
		{
			Transform mainController = GetMainController().transform;
			myAnchorPosition = mainController.position;
			myAnchorRotation = mainController.rotation;
			myVelocity = Vector3.zero;
			myTargetTime = -1.0f;
			myClock = 0.0f;
		}
		
		private void Update()
		{
			if (!myActive.val)
				return;
			
			Transform mainController = GetMainController().transform;
			
			myClock += Time.deltaTime;
			if (myClock > myTargetTime * 0.7f)
			{
				myClock = 0.0f;
				myTargetPosition = myAnchorPosition + Random.onUnitSphere * myRadius.val;
				myTargetRotation = Quaternion.Slerp(myAnchorRotation, Random.rotation, myRotationStrength.val);				
				
				float d = Vector3.Distance(mainController.position, myTargetPosition);
				d *= 100.0f; // centimeters				
				myTargetTime = Random.Range(d / mySpeedMin.val, d / mySpeedMax.val);
			}
			
			Vector3 pos = Vector3.SmoothDamp(mainController.position, myTargetPosition, ref myVelocity, myTargetTime);
			Quaternion rot =  Quaternion.Slerp(mainController.rotation, myTargetRotation, Time.deltaTime / myTargetTime);			
			mainController.SetPositionAndRotation(pos, rot);
		}
		
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				{
					JSONClass anchor = new JSONClass();
					JSONClass pos = new JSONClass();
					pos["x"].AsFloat = myAnchorPosition.x;
					pos["y"].AsFloat = myAnchorPosition.y;
					pos["z"].AsFloat = myAnchorPosition.z;
					anchor["position"] = pos;	
					JSONClass rot = new JSONClass();
					Vector3 euler = myAnchorRotation.eulerAngles;
					rot["x"].AsFloat = euler.x;
					rot["y"].AsFloat = euler.y;
					rot["z"].AsFloat = euler.z;
					anchor["rotation"] = rot;
					jc["anchor"] = anchor;
				}
				{
					JSONClass velocity = new JSONClass();
					velocity["x"].AsFloat = myVelocity.x;
					velocity["y"].AsFloat = myVelocity.y;
					velocity["z"].AsFloat = myVelocity.z;
					jc["velocity"] = velocity;
				}
			}
			return jc;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
			
			if (!physicalLocked && restorePhysical)
			{
				ResetAnchor();
				if (jc.HasKey("anchor"))
				{
					JSONClass anchor = jc["anchor"].AsObject;				
					if (anchor.HasKey("position"))
					{
						JSONClass pos = anchor["position"].AsObject;
						myAnchorPosition.x = pos["x"].AsFloat;
						myAnchorPosition.y = pos["y"].AsFloat;
						myAnchorPosition.z = pos["z"].AsFloat;
					}
					if (anchor.HasKey("rotation"))
					{
						JSONClass rot = anchor["rotation"].AsObject;
						Vector3 euler;
						euler.x = rot["x"].AsFloat;
						euler.y = rot["y"].AsFloat;
						euler.z = rot["z"].AsFloat;
						myAnchorRotation.eulerAngles = euler;
					}
				}
				if (jc.HasKey("velocity"))
				{
					JSONClass velocity = jc["velocity"].AsObject;		
					myVelocity.x = velocity["x"].AsFloat;
					myVelocity.y = velocity["y"].AsFloat;
					myVelocity.z = velocity["z"].AsFloat;					
				}
			}
		}
	}
}