using System;
using System.Collections.Generic;
using UnityEngine;


namespace MacGruber
{
	[DefaultExecutionOrder(200)]
	public class ScreenFade : MVRScript
	{
		private GameObject myWindowCamera;
		private GameObject myMonitorRig;
		private Mesh myMesh;
		private Material myMaterial;
		private Color myColor = Color.black;
		private float myFadeClock = 0.0f;
		private JSONStorableBool myActive;
		private JSONStorableFloat myTime;
		
		public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ScreenFade</b></size>\nLogicBrick that can fade your scene on the screen to the specified color. Among other things, helpful for nice scene transitions, covering up loading sequences.\n\n" + 
				"<b>Notes:</b>\n" + 
				"- Does NOT affect VaM's main UI.\n" + 
				"- Does affect UI placed in the scene, e.g. UIButton atom.\n" + 
				"- Does affect the green/red target gizmos in edit mode, making it hard to select things by clicking them. Use the selection menu instead.\n" + 
				"- Can only affect only a single camera. If the WindowCamera is active, it applies to that.</color>",
				600.0f, true
			);
			
			
			GameObject prefab = GameObject.CreatePrimitive(PrimitiveType.Cube);
			myMesh = prefab.GetComponent<MeshFilter>().mesh;
			DestroyImmediate(prefab);
			
			Shader[] shaders = Resources.FindObjectsOfTypeAll<Shader>();
			Shader shader = Array.Find(shaders, s => s.name == "Custom/TransparentHUD");
			myMaterial = new Material(shader);
			myMaterial.renderQueue = 10000;
			
			Atom windowCameraAtom = GetSceneAtoms().Find(a => a.type == "WindowCamera");
			CameraControl windowCamera = windowCameraAtom?.GetStorableByID("CameraControl") as CameraControl;
			myWindowCamera = windowCamera?.cameraGroup.gameObject;
			
			myMonitorRig = SuperController.singleton.MonitorRig.gameObject;

			myActive = Utils.SetupToggle(this, "Active", false, false);
			myTime = Utils.SetupSliderFloatWithRange(this, "Time (sec)", 1, 0, 10, false);
			myTime.constrained = false;
			
			JSONStorableColor colorUI = Utils.SetupColor(this, "Color", myColor, false);
			colorUI.setCallbackFunction += (float h, float s, float v) => {	myColor = HSVColorPicker.HSVToRGB(h, s, v);	};
		}
		
		private void OnDestroy()
		{
			Destroy(myMaterial);
			myMaterial = null;
			myMesh = null;
		}
				
		private void LateUpdate()
		{			
			float delta = myActive.val ? Time.deltaTime: -Time.deltaTime;
			delta /= Mathf.Max(myTime.val, 0.001f);
			myFadeClock = Mathf.Clamp01(myFadeClock+delta);
			if (myFadeClock > 0.0f)
			{
				myColor.a = Mathf.SmoothStep(0.0f, 1.0f, myFadeClock);
				myMaterial.color = myColor;			
				
				bool usingWindowCamera = myWindowCamera != null && myWindowCamera.activeSelf &&	!myMonitorRig.activeSelf;
				Transform transform = usingWindowCamera ? myWindowCamera.transform : CameraTarget.centerTarget.transform;
				Graphics.DrawMesh(myMesh, transform.localToWorldMatrix, myMaterial, gameObject.layer, null, 0, null, castShadows: false, receiveShadows: false);
			}
		}
	}
}