﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class SequenceChoice : MVRScript
    {
		private JSONStorableString myCurrentItemInfo;
		private JSONStorableStringChooser myItemChooser;
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableBool myLoop;
		private JSONStorableBool myRandomShuffle;
		private JSONStorableString myItemName;
		private JSONStorableString mySelectItem;

		private List<object> myItemUI = new List<object>();

		private class Item
		{
			public Item(MVRScript script, string name)
			{
				myName = name;
				Index = -1;
				Trigger = new EventTrigger(script, "On" + name);
			}
			
			public Item(MVRScript script, string name, Item source)
			{
				Index = -1;
				Trigger = new EventTrigger(source.Trigger);
				Name = name;
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					Trigger.Name = "On" + value;
				}
			}
			
			public int Index { get; set; }

			public EventTrigger Trigger;
			private string myName;
		}

		private List<Item> myItems = new List<Item>();
		private List<int> myIndices = new List<int>();
		private int mySelectedIndex = 0;
		private bool myNeedUIRefresh = false;

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SequenceChoice</b></size>\nLogicBrick similar to <i>RandomChoice</i> and <i>SelectChoice</i>. Each time when <i>Trigger</i> is called, it executes the next <i>Item</i> in the sequence.\n\n" +
				"Enable <i>Use Loop</i> to automatically restart at the beginning when reaching the end of the Sequence. You can also call <i>Reset</i> at any time.\n\n" + 
				"Enable <i>Use Random Shuffle</i> to shuffle the items in the sequence for each full loop through them.</color>",
				595.0f, true
			);
			
			Utils.SetupSpacer(this, 15, true);
			Utils.SetupInfoOneLine(this, "<size=30><b>Debug</b></size>", true);
			myCurrentItemInfo = new JSONStorableString("CurrentItem", "Current Item:");	
			Utils.SetupInfoOneLine(this, myCurrentItemInfo, true);
			
			myItemChooser = new JSONStorableStringChooser("Item", new List<string>(), "", "Item");		
			myItemChooser.isStorable = myItemChooser.isRestorable = false;
			myItemChooser.setCallbackFunction += UISetItem;
			CreateScrollablePopup(myItemChooser, false);

			Utils.SetupTwinButton(this, "Add Item", UIAddItem, "Duplicate Item", UIDuplicateItem, false);
			Utils.SetupButton(this, "Remove Item", UIRemoveItem, false);
			
			Utils.SetupSpacer(this, 5, false);
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);
			myLoop = Utils.SetupToggle(this, "Use Loop", true, false);
			myRandomShuffle = Utils.SetupToggle(this, "Use Random Shuffle", false, false);
			myRandomShuffle.setCallbackFunction += (bool v) => { RebuildIndices(); };
			
			Utils.SetupSpacer(this, 5, false);

			Utils.SetupButton(this, "Select Item", SelectItemButton, false);
			Utils.SetupButton(this, "Test Trigger", TriggerAction, false);
			
			mySelectItem = new JSONStorableString("SelectItem", "", SelectItemAction);
			mySelectItem.isStorable = mySelectItem.isRestorable = false;
			RegisterString(mySelectItem);
			Utils.SetupAction(this, "Reset", ResetItemAction);
			Utils.SetupAction(this, "Trigger", TriggerAction);

			Utils.SetupSpacer(this, 15, false);

			myItemName = new JSONStorableString("Name", "", UIRenameItem);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}
		
		private void SelectItemButton()
		{			
			int actualIndex = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (actualIndex >= 0)
				SelectItem(myItems[actualIndex].Index);
			else
				ResolveItem();
		}
		
		private void SelectItemAction(string newItem)
		{
			mySelectItem.valNoCallback = string.Empty;
			int actualIndex = myItemChooser.choices.IndexOf(newItem);
			if (actualIndex >= 0)
			{
				SelectItem(myItems[actualIndex].Index);
			}
			else
			{
				SuperController.LogError($"LogicBricks.SelectItem: Unknown item '{newItem}'.");
				ResolveItem();
			}
		}
		
		private void ResetItemAction()
		{
			if (myRandomShuffle.val)
				RebuildIndices();			
			SelectItem(0);
		}
		
		private void SelectItem(int newItemIndex)
		{
			if (newItemIndex < 0)
				mySelectedIndex = 0;
			else if (newItemIndex >= myItems.Count)
				mySelectedIndex = myLoop.val ? 0 : myItems.Count;
			else
				mySelectedIndex = newItemIndex;
			
			if (mySelectedIndex >= myItems.Count)
				myCurrentItemInfo.val = "Current Item:";
			else
				myCurrentItemInfo.val = "Current Item: " + myItems[myIndices[mySelectedIndex]].Name;
		}
		
		private void ResolveItem()
		{
			SelectItem(mySelectedIndex);
		}
		
		private void UISetItem(string v)
		{
			Utils.RemoveUIElements(this, myItemUI);

			int idx = myItemChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			Item item = myItems[idx];

			UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Item</b></size>", false);
			myItemUI.Add(optionHeader);

			myItemName.valNoCallback = item.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myItemName, false);
			myItemUI.Add(nameInput);

			UIDynamicTwinButton moveUpButtons = Utils.SetupTwinButton(this, "Move to Top", UIMoveTopItem, "Move Up", UIMoveUpItem, false);
			myItemUI.Add(moveUpButtons);
			
			UIDynamicTwinButton moveDownButtons = Utils.SetupTwinButton(this, "Move to Bottom", UIMoveBottomItem, "Move Down", UIMoveDownItem, false);
			myItemUI.Add(moveDownButtons);

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + item.Trigger.Name, item.Trigger.OpenPanel, false);
			myItemUI.Add(actionsButton);
		}
		
		private void UIMoveTopItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx >= 1)
				UISwapItem(idx, 0);
		}
		
		private void UIMoveUpItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx >= 1)
				UISwapItem(idx, idx-1);
		}
		
		private void UIMoveBottomItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx < myItemChooser.choices.Count - 1)
				UISwapItem(idx, myItemChooser.choices.Count - 1);
		}
		
		private void UIMoveDownItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx < myItemChooser.choices.Count - 1)
				UISwapItem(idx, idx+1);
		}
		
		private void UISwapItem(int oldIdx, int newIdx)
		{
			Item tmpItem = myItems[newIdx];
			myItems[newIdx] = myItems[oldIdx];
			myItems[oldIdx] = tmpItem;

			List<string> options = myItemChooser.choices;
			string tmpName = options[newIdx];
			options[newIdx] = options[oldIdx];
			options[oldIdx] = tmpName;
			myItemChooser.choices = null; // force UI sync
			myItemChooser.choices = options;
			myItemChooser.valNoCallback = options[oldIdx];
			myItemChooser.valNoCallback = options[newIdx];
			
			for (int i=0; i<myIndices.Count; ++i)
				myItems[myIndices[i]].Index = i;
			
			ResolveItem();
		}
		
		private void UIAddItem()
		{
			UIAddItem(null);
		}

		private void UIAddItem(Item source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Item" + i;
				if (!myItemChooser.choices.Contains(name))
				{
					Item item;
					if (source == null)
						item = new Item(this, name);
					else
						item = new Item(this, name, source);
					myItems.Add(item);
					List<string> options = myItemChooser.choices;
					options.Clear();
					for (int e=0; e<myItems.Count; ++e)
						options.Add(myItems[e].Name);
					myItemChooser.choices = null; // force UI sync
					myItemChooser.choices = options;
					myItemChooser.val = name;
					
					RebuildIndices();
					ResolveItem();
					return;
				}
			}
			SuperController.LogError("Sequence: Too many items!");
		}
		
		private void UIDuplicateItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			Item item = idx >= 0 ? myItems[idx] : null;
			UIAddItem(item);
		}

		private void UIRemoveItem()
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx < 0)
				return;

			List<string> options = myItemChooser.choices;
			options.RemoveAt(idx);
			myItems[idx].Trigger.Remove();
			myItems.RemoveAt(idx);
			myItemChooser.choices = null; // force UI sync
			myItemChooser.choices = options;

			if (idx < options.Count)
				myItemChooser.val = options[idx];
			else if (options.Count > 0)
				myItemChooser.val = options[options.Count-1];
			else
				myItemChooser.val = "";
			
			RebuildIndices();
			ResolveItem();
		}

		private void UIRenameItem(string name)
		{
			int idx = myItemChooser.choices.IndexOf(myItemChooser.val);
			if (idx < 0)
				return;

			myItems[idx].Name = name;

			List<string> options = myItemChooser.choices;
			options[idx] = name;
			myItemChooser.choices = null; // force UI sync
			myItemChooser.choices = options;
			myItemChooser.val = name;
			
			ResolveItem();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myItems.Count; ++i)
				myItems[i].Trigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				
				jc["Selected"].AsInt = mySelectedIndex;

				JSONArray clist = new JSONArray();
				for (int i=0; i<myItems.Count; ++i)
				{
					Item item = myItems[i];
					JSONClass cc = new JSONClass();
					cc["Name"] = item.Name;
					cc[item.Trigger.Name] = item.Trigger.GetJSON(base.subScenePrefix);
					clist.Add("", cc);
				}
				jc["Items"] = clist;
				
				JSONArray ilist = new JSONArray();
				for (int i=0; i<myIndices.Count; ++i)
				{
					ilist[i].AsInt = myIndices[i];
				}
				
				jc["Indices"] = ilist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myItems.Count; ++i)
					myItems[i].Trigger.Remove();
				myItems.Clear();
				myIndices.Clear();

				JSONArray clist = jc["Items"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass cc = clist[i].AsObject;
					Item item = new Item(this, cc["Name"]);
					item.Trigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myItems.Add(item);					
				}
				
				bool dataError = false;
				JSONArray ilist = jc["Indices"].AsArray;
				for (int i=0; i<ilist.Count; ++i)
				{
					int index = ilist[i].AsInt;
					if (index >= 0 && index < myItems.Count)
					{
						myIndices.Add(index);
						myItems[index].Index = i;					
					}
					else
					{
						dataError = true;
					}
				}
				for (int i=0; i<myItems.Count; ++i)
				{
					dataError |= myItems[i].Index < 0;
				}
				dataError |= myIndices.Count != myItems.Count;
				if (dataError) // someone messed with the indices by hand....need to repair
				{
					RebuildIndices();
					SuperController.LogError("LogicBricks Sequence plugin found broken index data. Please resave the scene to fix.");
				}
				
				List<string> options = myItemChooser.choices;
				options.Clear();
				for (int e=0; e<myItems.Count; ++e)
					options.Add(myItems[e].Name);
				myItemChooser.choices = null; // force UI sync
				myItemChooser.choices = options;

				if (options.Count == 0)
					myItemChooser.valNoCallback = "";
				else if (!options.Contains(myItemChooser.val))
					myItemChooser.valNoCallback = options[0];
				
				SelectItem(jc["Selected"].AsInt);		
				
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myItems.Count; ++i)
				myItems[i].Trigger.Update();
			
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myItemChooser.setCallbackFunction(myItemChooser.val);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myItems.Count; ++i)
				myItems[i].Trigger.Remove();
			myItems.Clear();

			Utils.OnDestroyUI();
		}

		private void TriggerAction()
		{
			if (!myAcceptTrigger.val || myItems.Count == 0)
				return;
			
			if (mySelectedIndex < 0 || (myLoop.val && mySelectedIndex >= myItems.Count))
			{
				mySelectedIndex = 0;
				if (myLoop.val && myRandomShuffle.val)
					RebuildIndices();
			}
			
			if (mySelectedIndex >= 0 && mySelectedIndex < myItems.Count)
				myItems[myIndices[mySelectedIndex++]].Trigger.Trigger();
			
			if (mySelectedIndex >= myItems.Count && myLoop.val && myRandomShuffle.val)
				RebuildIndices();
			
			ResolveItem();
		}
				
		private void RebuildIndices()
		{
			Item currentItem = null;
			if (mySelectedIndex >= 0 && mySelectedIndex < myIndices.Count)
				currentItem = myItems[myIndices[mySelectedIndex]];
			
			myIndices.Clear();
			for (int i=0; i<myItems.Count; ++i)
				myIndices.Add(i);

			if (myRandomShuffle.val)
			{
				int n = myIndices.Count;  
				while (n > 1)
				{  
					--n;
					int k = Random.Range(0, n);
					int tmp = myIndices[k];  
					myIndices[k] = myIndices[n];  
					myIndices[n] = tmp;  
				}
			}
			
			for (int i=0; i<myIndices.Count; ++i)
				myItems[myIndices[i]].Index = i;

			mySelectedIndex = currentItem != null ? currentItem.Index : -1;
		}
    }
}
