﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class SequenceMachine : MVRScript
    {
		private JSONStorableString myCurrentStateInfo;
		private JSONStorableString myCurrentClockInfo;
		private JSONStorableStringChooser myStateChooser;
		private JSONStorableBool myLoop;
		private JSONStorableBool myRandomShuffle;
		private JSONStorableBool myPaused;
		private JSONStorableString myStateName;
		private JSONStorableBool myStateWaitForSync;
		private JSONStorableBool myStateInfiniteDuration;
		private JSONStorableFloat myStateMinDuration;
		private JSONStorableFloat myStateMaxDuration;
		private JSONStorableString mySwitchState;

		private List<object> myStateUI = new List<object>();

		private class State
		{
			public State(MVRScript script, string name)
			{
				myName = name;
				Index = -1;
				EnterTrigger = new EventTrigger(script, "OnEnter" + name);
				ValueTrigger = new FloatTrigger(script, "OnValue" + name);
				ExitTrigger = new EventTrigger(script, "OnExit" + name);
			}
			
			public State(MVRScript script, string name, State source)
			{
				Index = -1;
				EnterTrigger = new EventTrigger(source.EnterTrigger);
				ValueTrigger = new FloatTrigger(source.ValueTrigger);
				ExitTrigger = new EventTrigger(source.ExitTrigger);
				Name = name;
				InfiniteDuration = source.InfiniteDuration;
				WaitForSync = source.WaitForSync;
				MinDuration = source.MinDuration;
				MaxDuration = source.MaxDuration;
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					EnterTrigger.Name = "OnEnter" + value;
					ValueTrigger.Name = "OnValue" + value;
					ExitTrigger.Name = "OnExit" + value;
				}
			}
			
			public int Index { get; set; }

			public EventTrigger EnterTrigger;
			public FloatTrigger ValueTrigger;
			public EventTrigger ExitTrigger;
			private string myName;
			public bool InfiniteDuration = false;
			public bool WaitForSync = false;
			public float MinDuration = 3.0f;
			public float MaxDuration = 5.0f;
		}

		private List<State> myStates = new List<State>();
		private List<int> myIndices = new List<int>();
		private State myCurrentState = null;
		private int myCurrentIndex = 0;
		private float myClock = 0.0f;
		private float myDuration = 1.0f;
		private float myValue = -1.0f;
		private int myClockInt = -1;
		private bool myNeedUIRefresh = false;

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SequenceMachine</b></size>\nLogicBrick similar to <i>StateMachine</i>, except transitions are fixed so states are gone through in sequence.\n\n" + 
				"You can reorder states in the sequence using the <i>Move Up/Down/Top/Bottom</i> buttons.\n\n" +
				"Enable <i>Use Loop</i> to automatically restart at the beginning when reaching the end of the Sequence. You can also call <i>Reset</i> at any time.\n\n" + 
				"Enable <i>Use Random Shuffle</i> to shuffle the states in the sequence for each full loop through them.</color>",
				730.0f, true
			);
			
			Utils.SetupSpacer(this, 5, true);
			Utils.SetupInfoOneLine(this, "<size=30><b>Debug</b></size>", true);
			myCurrentStateInfo = new JSONStorableString("CurrentState", "Current State:");
			myCurrentClockInfo = new JSONStorableString("CurrentClock", "Current Clock:");
			Utils.SetupInfoOneLine(this, myCurrentStateInfo, true);
			Utils.SetupInfoOneLine(this, myCurrentClockInfo, true);
			Utils.SetupSpacer(this, 5, true);
			
			myStateChooser = new JSONStorableStringChooser("State", new List<string>(), "", "State");		
			myStateChooser.isStorable = myStateChooser.isRestorable = false;
			myStateChooser.setCallbackFunction += UISetState;
			CreateScrollablePopup(myStateChooser, false);

			Utils.SetupTwinButton(this, "Add State", UIAddState, "Duplicate State", UIDuplicateState, false);
			Utils.SetupButton(this, "Remove State", UIRemoveState, false);
			
			Utils.SetupSpacer(this, 5, false);
						
			Utils.SetupTwinButton(this, "Switch State", UISwitchState, "Skip State", UISkipState, false);
			
			myPaused = new JSONStorableBool("Time Paused (not saved)", false);
			CreateToggle(myPaused, false);
			myLoop = Utils.SetupToggle(this, "Use Loop", true, false);
			myRandomShuffle = Utils.SetupToggle(this, "Use Random Shuffle", false, false);
			myRandomShuffle.setCallbackFunction += (bool v) => { RebuildIndices(); };
			
			mySwitchState = new JSONStorableString("SwitchState", "", SwitchStateAction);
			mySwitchState.isStorable = mySwitchState.isRestorable = false;
			RegisterString(mySwitchState);
			Utils.SetupAction(this, "Reset", ResetStateAction);
			Utils.SetupAction(this, "TriggerSync", TriggerSyncAction);

			Utils.SetupSpacer(this, 0, false);

			myStateName = new JSONStorableString("Name", "", UIRenameState);
			myStateInfiniteDuration = new JSONStorableBool("Infinite Duration", false);
			myStateInfiniteDuration.setCallbackFunction += UISetInfiniteDuration;
			myStateWaitForSync = new JSONStorableBool("Wait for TriggerSync", false);
			myStateWaitForSync.setCallbackFunction += UISetWaitForSync;
			myStateMinDuration = new JSONStorableFloat("Min Duration (sec)", 3, 0, 10, true, true);
			myStateMinDuration.setCallbackFunction += UISetMinDuration;
			myStateMinDuration.constrained = false;
			myStateMaxDuration = new JSONStorableFloat("Max Duration (sec)", 5, 0, 10, true, true);
			myStateMaxDuration.setCallbackFunction += UISetMaxDuration;
			myStateMaxDuration.constrained = false;
			
			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}
		
		private void UISwitchState()
		{			
			int actualIndex = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (actualIndex >= 0)
				SwitchState(myStates[actualIndex].Index);
		}
		
		private void UISkipState()
		{
			myClock = 0.0f;
		}
		
		private void SwitchStateAction(string newState)
		{
			mySwitchState.valNoCallback = string.Empty;
			int actualIndex = myStateChooser.choices.IndexOf(newState);
			if (actualIndex >= 0)
				SwitchState(myStates[actualIndex].Index);
			else
				SuperController.LogError($"LogicBricks.SwitchState: Unknown state '{newState}'.");
		}
		
		private void ResetStateAction()
		{
			if (myRandomShuffle.val)
				RebuildIndices();			
			SwitchState(0);
		}
		
		private void TriggerSyncAction()
		{
			if (myCurrentState == null || !myCurrentState.WaitForSync)
				return;
			if (myClock <= 0.0f && !myCurrentState.InfiniteDuration)
				SwitchToNextState();
		}
		
		private void SwitchState(int newStateIndex)
		{
			if (myCurrentState != null)
			{
				if (myValue != 1.0f)
					myCurrentState.ValueTrigger.Trigger(1.0f);
				myCurrentState.ExitTrigger.Trigger();
				myCurrentState = null;
				myValue = -1.0f;
			}
			
			if (newStateIndex < 0)
			{
				myCurrentIndex = 0;
			}
			else if (newStateIndex >= myStates.Count)
			{
				if (myRandomShuffle.val)
					RebuildIndices();
				myCurrentIndex = myLoop.val ? 0 : myStates.Count;
			}
			else
			{
				myCurrentIndex = newStateIndex;
			}
			
			if (myCurrentIndex >= 0 && myCurrentIndex < myStates.Count)
			{
				myCurrentState = myStates[myIndices[myCurrentIndex]];
				if (myCurrentState.InfiniteDuration)
					myClock = myDuration = 0.0f;
				else
					myClock = myDuration = UnityEngine.Random.Range(myCurrentState.MinDuration, myCurrentState.MaxDuration);
				myClockInt = -1;				
				myValue = 0.0f;
				
				myCurrentState.EnterTrigger.Trigger();
				myCurrentState.ValueTrigger.Trigger(0.0f);
			}
			
			UpdateCurrentStateInfo();
		}
		
		private void SwitchToNextState()
		{
			SwitchState(myCurrentIndex+1);
		}
				
		private void UISetState(string v)
		{
			Utils.RemoveUIElements(this, myStateUI);

			int idx = myStateChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			State state = myStates[idx];

			UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>State</b></size>", false);
			myStateUI.Add(optionHeader);

			myStateName.valNoCallback = state.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myStateName, false);
			myStateUI.Add(nameInput);

			UIDynamicTwinButton moveUpButtons = Utils.SetupTwinButton(this, "Move to Top", UIMoveTopState, "Move Up", UIMoveUpState, false);
			myStateUI.Add(moveUpButtons);
			
			UIDynamicTwinButton moveDownButtons = Utils.SetupTwinButton(this, "Move to Bottom", UIMoveBottomState, "Move Down", UIMoveDownState, false);
			myStateUI.Add(moveDownButtons);
			
			myStateInfiniteDuration.valNoCallback = state.InfiniteDuration;
			CreateToggle(myStateInfiniteDuration, false);
			myStateUI.Add(myStateInfiniteDuration);

			if (!state.InfiniteDuration)
			{
				myStateWaitForSync.valNoCallback = state.WaitForSync;
				CreateToggle(myStateWaitForSync, false);
				myStateUI.Add(myStateWaitForSync);

				myStateMinDuration.valNoCallback = state.MinDuration;
				myStateMaxDuration.valNoCallback = state.MaxDuration;
				Utils.AdjustSliderRange(myStateMinDuration);
				Utils.AdjustSliderRange(myStateMaxDuration);
				UIDynamicSlider sliderMin = CreateSlider(myStateMinDuration, false);
				UIDynamicSlider sliderMax = CreateSlider(myStateMaxDuration, false);
				sliderMin.rangeAdjustEnabled = true;
				sliderMax.rangeAdjustEnabled = true;
				myStateUI.Add(myStateMinDuration);
				myStateUI.Add(myStateMaxDuration);
			}
			
			UIDynamicTextInfo triggersHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Triggers</b></size>", true);
			myStateUI.Add(triggersHeader);
			UIDynamicButton enterButton = Utils.SetupButton(this, "Actions " + state.EnterTrigger.Name, state.EnterTrigger.OpenPanel, true);
			UIDynamicButton valueButton = Utils.SetupButton(this, "Actions " + state.ValueTrigger.Name, state.ValueTrigger.OpenPanel, true);
			UIDynamicButton exitButton = Utils.SetupButton(this, "Actions " + state.ExitTrigger.Name, state.ExitTrigger.OpenPanel, true);
			myStateUI.Add(enterButton);
			myStateUI.Add(valueButton);
			myStateUI.Add(exitButton);
			
		}
		
		private void UIMoveTopState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx >= 1)
				UISwapState(idx, 0);
		}
		
		private void UIMoveUpState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx >= 1)
				UISwapState(idx, idx-1);
		}
		
		private void UIMoveBottomState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < myStateChooser.choices.Count - 1)
				UISwapState(idx, myStateChooser.choices.Count - 1);
		}
		
		private void UIMoveDownState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < myStateChooser.choices.Count - 1)
				UISwapState(idx, idx+1);
		}
		
		private void UISwapState(int oldIdx, int newIdx)
		{
			State tmpState = myStates[newIdx];
			myStates[newIdx] = myStates[oldIdx];
			myStates[oldIdx] = tmpState;

			List<string> options = myStateChooser.choices;
			string tmpName = options[newIdx];
			options[newIdx] = options[oldIdx];
			options[oldIdx] = tmpName;
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;
			myStateChooser.valNoCallback = options[oldIdx];
			myStateChooser.valNoCallback = options[newIdx];
			
			for (int i=0; i<myIndices.Count; ++i)
				myStates[myIndices[i]].Index = i;
		}
		
		private void UIAddState()
		{
			UIAddState(null);
		}

		private void UIAddState(State source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "State" + i;
				if (!myStateChooser.choices.Contains(name))
				{
					State state;
					if (source == null)
						state = new State(this, name);
					else
						state = new State(this, name, source);
					myStates.Add(state);
					List<string> options = myStateChooser.choices;
					options.Clear();
					for (int e=0; e<myStates.Count; ++e)
						options.Add(myStates[e].Name);
					myStateChooser.choices = null; // force UI sync
					myStateChooser.choices = options;
					myStateChooser.val = name;
					
					RebuildIndices();
					return;
				}
			}
			SuperController.LogError("Sequence: Too many states!");
		}
		
		private void UIDuplicateState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			State state = idx >= 0 ? myStates[idx] : null;
			UIAddState(state);
		}

		private void UIRemoveState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < 0)
				return;

			List<string> options = myStateChooser.choices;
			options.RemoveAt(idx);			
			if (myCurrentState == myStates[idx])
			{
				myCurrentState = null;
				UpdateCurrentStateInfo();
			}
			myStates[idx].EnterTrigger.Remove();
			myStates[idx].ValueTrigger.Remove();
			myStates[idx].ExitTrigger.Remove();			
			myStates.RemoveAt(idx);
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;

			if (idx < options.Count)
				myStateChooser.val = options[idx];
			else if (options.Count > 0)
				myStateChooser.val = options[options.Count-1];
			else
				myStateChooser.val = "";
			
			RebuildIndices();
		}

		private void UIRenameState(string name)
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < 0)
				return;

			myStates[idx].Name = name;

			List<string> options = myStateChooser.choices;
			options[idx] = name;
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;
			myStateChooser.val = name;
			
			if (myStates[idx] == myCurrentState)
				UpdateCurrentStateInfo();
		}
		
		private State GetSelectedState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			return idx >= 0 ? myStates[idx] : null;
		}
		
		private void UISetInfiniteDuration(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.InfiniteDuration = v;
			state.WaitForSync &= !v;
			myNeedUIRefresh = true;

			myClockInt = -1;
			myValue = -1.0f;
			UpdateCurrentStateInfo();
			UpdateCurrentStateValue();
		}

		private void UISetWaitForSync(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.WaitForSync = v;
		}

		private void UISetMinDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MinDuration = v;
			Utils.AdjustMaxSliderFromMin(v, myStateMaxDuration);
			state.MaxDuration = myStateMaxDuration.val;
		}

		private void UISetMaxDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MaxDuration = v;
			state.MinDuration = Mathf.Min(v, state.MinDuration);
			myStateMinDuration.valNoCallback = state.MinDuration;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.SyncAtomNames();
				myStates[i].ValueTrigger.SyncAtomNames();
				myStates[i].ExitTrigger.SyncAtomNames();
			}
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				
				jc["Current"].AsInt = myCurrentIndex;
				jc["Clock"].AsFloat = myClock;
				jc["Duration"].AsFloat = myDuration;

				JSONArray clist = new JSONArray();
				for (int i=0; i<myStates.Count; ++i)
				{
					State state = myStates[i];
					JSONClass cc = new JSONClass();
					cc["Name"] = state.Name;
					cc["InfiniteDuration"].AsBool = state.InfiniteDuration;
					cc["WaitForSync"].AsBool = state.WaitForSync;
					cc["MinDuration"].AsFloat = state.MinDuration;
					cc["MaxDuration"].AsFloat = state.MaxDuration;
					cc[state.EnterTrigger.Name] = state.EnterTrigger.GetJSON(base.subScenePrefix);
					cc[state.ValueTrigger.Name] = state.ValueTrigger.GetJSON(base.subScenePrefix);
					cc[state.ExitTrigger.Name] = state.ExitTrigger.GetJSON(base.subScenePrefix);
					clist.Add("", cc);
				}
				jc["States"] = clist;
				
				JSONArray ilist = new JSONArray();
				for (int i=0; i<myIndices.Count; ++i)
				{
					ilist[i].AsInt = myIndices[i];
				}
				jc["Indices"] = ilist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myStates.Count; ++i)
				{
					myStates[i].EnterTrigger.Remove();
					myStates[i].ValueTrigger.Remove();
					myStates[i].ExitTrigger.Remove();
				}
				myStates.Clear();
				myIndices.Clear();
				myCurrentState = null;

				JSONArray clist = jc["States"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass cc = clist[i].AsObject;
					State state = new State(this, cc["Name"]);
					state.InfiniteDuration = cc["InfiniteDuration"].AsBool;
					state.WaitForSync = cc["WaitForSync"].AsBool && !state.InfiniteDuration;
					state.MinDuration = cc["MinDuration"].AsFloat;
					state.MaxDuration = cc["MaxDuration"].AsFloat;
					state.EnterTrigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					state.ValueTrigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					state.ExitTrigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myStates.Add(state);					
				}
				
				bool dataError = false;
				JSONArray ilist = jc["Indices"].AsArray;
				for (int i=0; i<ilist.Count; ++i)
				{
					int index = ilist[i].AsInt;
					if (index >= 0 && index < myStates.Count)
					{
						myIndices.Add(index);
						myStates[index].Index = i;					
					}
					else
					{
						dataError = true;
					}
				}
				for (int i=0; i<myStates.Count; ++i)
				{
					dataError |= myStates[i].Index < 0;
				}
				dataError |= myIndices.Count != myStates.Count;
				if (dataError) // someone messed with the indices by hand....need to repair
				{
					RebuildIndices();
					SuperController.LogError("LogicBricks Sequence plugin found broken index data. Please resave the scene to fix.");
				}
				
				List<string> options = myStateChooser.choices;
				options.Clear();
				for (int e=0; e<myStates.Count; ++e)
					options.Add(myStates[e].Name);
				myStateChooser.choices = null; // force UI sync
				myStateChooser.choices = options;

				if (options.Count == 0)
					myStateChooser.valNoCallback = "";
				else if (!options.Contains(myStateChooser.val))
					myStateChooser.valNoCallback = options[0];
				
				myClock = 0.0f;
				myDuration = 0.0f;
				myClockInt = -1;
				myValue = -1.0f;
				myCurrentIndex = jc["Current"].AsInt;
				if (myCurrentIndex >= 0 && myCurrentIndex < myStates.Count)
				{
					myCurrentState = myStates[myIndices[myCurrentIndex]];
					myClock = jc["Clock"].AsFloat;
					myDuration = jc["Duration"].AsFloat;
				}
				UpdateCurrentStateInfo();
				
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.Update();
				myStates[i].ValueTrigger.Update();
				myStates[i].ExitTrigger.Update();
			}
			
			bool uiActive = UITransform.gameObject.activeInHierarchy;
			if (uiActive && myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myStateChooser.setCallbackFunction(myStateChooser.val);
			}
			
			if (myCurrentState == null || myPaused.val)
				return;
			
			myClock = Mathf.Max(myClock - Time.deltaTime, 0.0f);
			if (myClock <= 0.0f && !myCurrentState.InfiniteDuration && !myCurrentState.WaitForSync )
			{
				SwitchToNextState();
			}
			else
			{
				if (uiActive)
					UpdateCurrentStateInfo();
				UpdateCurrentStateValue();
			}
		}

		private void UpdateCurrentStateInfo()
		{
			if (myCurrentState != null)
			{
				if (myClockInt < 0)
				{
					myCurrentStateInfo.val = "Current State: " + myCurrentState.Name;
				}

				int clockInt = Mathf.CeilToInt(myClock*10.0f);				
				if (myCurrentState.InfiniteDuration)
				{
					if (myClockInt < 0)
						myCurrentClockInfo.val = "Current Clock: Infinite";
				}
				else if (clockInt == 0 && myCurrentState.WaitForSync)
				{
					myCurrentClockInfo.val = "Current Clock: Waiting for TriggerSync";
				}
				else if (myClockInt < 0 || myClockInt != clockInt)
				{
					myCurrentClockInfo.val = string.Format("Current Clock: {0:F1} sec", clockInt/10.0f);
				}
				myClockInt = clockInt;
			}
			else
			{
				myCurrentStateInfo.val = "Current State:";
				myCurrentClockInfo.val = "Current Clock:";
			}
		}
		
		private void UpdateCurrentStateValue()
		{
			if (myCurrentState != null)
			{
				float v = 0.0f;
				if (!myCurrentState.InfiniteDuration)
					v = myDuration>0.001f ? 1.0f - Mathf.Clamp01(myClock / myDuration) : 1.0f;
				if (myValue != v)
					myCurrentState.ValueTrigger.Trigger(v);
				myValue = v;
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.Remove();
				myStates[i].ValueTrigger.Remove();
				myStates[i].ExitTrigger.Remove();
			}
			myStates.Clear();

			Utils.OnDestroyUI();
		}
				
		private void RebuildIndices()
		{
			myIndices.Clear();
			for (int i=0; i<myStates.Count; ++i)
				myIndices.Add(i);

			if (myRandomShuffle.val)
			{
				int n = myIndices.Count;  
				while (n > 1)
				{  
					int k = Random.Range(0, n);
					--n;
					int tmp = myIndices[k];  
					myIndices[k] = myIndices[n];  
					myIndices[n] = tmp;  
				}
			}
			
			for (int i=0; i<myIndices.Count; ++i)
				myStates[myIndices[i]].Index = i;			

			myCurrentIndex = myCurrentState != null ? myCurrentState.Index : -1;
		}
    }
}
